/* 
 * Main source file of the WOM_finder program.
 * All core functions are located in this file
 * 
 * For comments and questions, please contact Bert Dobbelaere <bert.o.dobbelaere[AT]telenet[DOT]be>
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#include "support.h" // I/O stuff and solution check


static u32 nsymbols1=0,  ///< Number of first write symbols
           nsymbols2=0,  ///< Number of second write symbols
           nbit=0, ///< Bit length of the code
           ncomb=0, ///< Number of code points i.e. 2**nbit
           first_write_threshold=0, ///< Maximum number of bits set for a first write symbol
           numLowHammingPatterns=0, ///< Number of first write symbols supported by first_write_threshold
           max_score=0, ///< Value that the score function will have if we found a solution.
           nsymbols=0, ///< Total number of different symbols in play for both writes
           randomweight=0, ///< Random weight that will be distributed over the first write symbols
           rebalance_count=0; ///< Number of iterations before we reassign the random weights

/**
 * Keeps the set of 1st write patterns that can be altered into the current
 */
struct Precursor_t
{
    u32 nprecursors;
    u32 *indices;
};

static Precursor_t *precursors=NULL; ///< Precursors for each code point

/**
 * Helper class that keeps track of the number of 2nd write symbols that are reachable from a candidate 1st write
 * pattern code point. Symbols can be added more than once. A battery of coverage counters takes care of that.
 */
class Coveragecounters
{
    public:
        static void SetDim(u32 d); ///< Set number of counters per instance
        Coveragecounters();
        ~Coveragecounters();
        void addsymbol(u32 s); ///< Add a symbol to the set
        void removesymbol(u32 s); ///< Remove a symbol from the set
        u32 getcoverage() const; ///< Retrieve the number of 2nd write symbols that have a coverage >=1.
        void clear(); ///< Set 0 coverage state
        
    private:
        static u32 dim;
        u32 coverage;
        u32 *counts;
};

u32 Coveragecounters::dim=0;

void Coveragecounters::SetDim(u32 d)
{
    dim=d;
}

Coveragecounters::Coveragecounters()
{
    counts=new u32[dim];
    clear();
}

Coveragecounters::~Coveragecounters()
{
    delete [](counts);
}

void Coveragecounters::clear()
{
    memset(counts,0,dim*sizeof(u32));
    coverage=0;
}

void Coveragecounters::addsymbol(u32 s)
{
    if(s<nsymbols2)
    {
        counts[s]++;
        if(counts[s]==1)
            coverage++;
    }
}

void Coveragecounters::removesymbol(u32 s)
{
    if(s<nsymbols2)
    {
        assert(counts[s]>0);
        counts[s]--;
        if(counts[s]==0)
            coverage--;
    }
}

u32 Coveragecounters::getcoverage() const
{
    assert(coverage<=nsymbols2);
    return coverage;
}

static Coveragecounters *coveragecounters=NULL;


static u32 *idx2_1st_symbol_pattern=NULL; // To be used to look up the patterns with low hamming count
static u32 *pattern2_1st_symbol=NULL;
static u32 *code_table=NULL; // The sequence we'll try to transform into a solution

#ifndef EQUAL_WEIGHTS
static u32 *weights=NULL; // Modifying set of weight factors in an attempt to keep fluidity longer

void setWeights(u32 nsymbols1, u32 *weigths)
{
    for(u32 i=0;i<nsymbols1;i++)
    {
        weigths[i]=1; // Minimum score
    }
    
    for(u32 i=0;i<randomweight;i++)
    {
        weigths[ rand()%nsymbols1 ]++; // Distribute random weights over the 1st write symbols
        //TODO for randomweight>>nsymbols1, there are faster ways to do this
    }   
}
#endif

/**
 * This is the core heuristic function to assess how well a given code table is on its way to a solution. 
 * In its basic form, it determines how many (1st write symbol, 2nd write symbol) pairs can be supported
 * by the evaluated code table. The higher that figure, the more likely it is we are on the right track
 * to a solution. Although this metric is far from a perfect measure, it is easy to determine and has
 * shown good results in practice.
 * As an improvement to the original algorithm, the 1st write symbols can each have a different weight in
 * the score. Repeatedly reassigning these weights over time helps the algorithm to escape from local
 * maxima.
 * This routine performs the full score computation 'from scratch'. The two functions that follow
 * increase performance by updating the score in an incremental way.
 */
u32 computeCodeTableScore(const u32 code_table[], u32 numLowHammingPatterns)
{
    // Clear all coverage counters for each candidate 1st write pattern
    for(u32 c=0;c<numLowHammingPatterns;c++)
    {
        coveragecounters[c].clear();
    }
    
    // For each code point in the table, add its current symbol to the covered set of all precursors
    for(u32 p=0;p<ncomb;p++)
    {
        for(u32 idx=0;idx<precursors[p].nprecursors;idx++)
        {
            u32 c=precursors[p].indices[idx];
            coveragecounters[c].addsymbol( code_table[p] );
        }
    }
    
    // Now determine per 1st write symbol the candidate with the best coverage
    u32 max_symbol_score[nsymbols]={0};
    u32 rv=0;
    for(u32 c=0;c<numLowHammingPatterns;c++)
    {
        u32 pos=idx2_1st_symbol_pattern[c];
        u32 score=coveragecounters[c].getcoverage();
        if(score>max_symbol_score[code_table[pos]])
            max_symbol_score[code_table[pos]] = score;
    }
    
    // Compute the (weighted) total
    for(u32 i=0;i<nsymbols1;i++)
    {
#ifndef EQUAL_WEIGHTS
        rv+=weights[i]*max_symbol_score[i];
#else
        rv+=max_symbol_score[i];
#endif
    }
    return rv;
}

/**
 * Updates the coverage counters after replacing a symbol.
 * It is sufficient to update only the precursor list for the code point
 */
void updateCodeTableCoverage(const u32 code_table[], u32 pos, u32 oldsymbol)
{   
    for(u32 idx=0;idx<precursors[pos].nprecursors;idx++)
    {
        u32 c=precursors[pos].indices[idx];
        coveragecounters[c].removesymbol( oldsymbol);
        coveragecounters[c].addsymbol( code_table[pos] );
    }
}

/**
 * Computes an updated score when one or more code points were changed.
 * The coverage counters need to be updated before calling this function.
 */
u32 updateCodeTableScore(const u32 code_table[], u32 numLowHammingPatterns)
{       
    u32 max_symbol_score[nsymbols]={0};
    u32 rv=0;
    for(u32 c=0;c<numLowHammingPatterns;c++)
    {
        u32 pos=idx2_1st_symbol_pattern[c];
        u32 score=coveragecounters[c].getcoverage();
        if(score>max_symbol_score[code_table[pos]])
            max_symbol_score[code_table[pos]] = score;
    }
    for(u32 i=0;i<nsymbols1;i++)
    {
#ifndef EQUAL_WEIGHTS
        rv+=weights[i]*max_symbol_score[i];
#else
        rv+=max_symbol_score[i];
#endif
    }
    assert(rv<=max_score);
    return rv;
}



/**
 * Find a solution code table, based on a random hill climbing method. For this function,
 * failure is not an option. It will keep trying until a solution was found, which is in 
 * many cases forever.
 * Apart from the fast iteration loop the weights of the score function are modified as a
 * less frequent task, in an attempt to escape from local maxima
 */

void findsolution(u32 code_table[], u32 numLowHammingPatterns)
{
    // Load a previously stored intermediate state if it exists
    bool isrestored = attemptRestoreState(nsymbols1, nsymbols2, nbit, code_table);
    if(!isrestored)
    {
        // Set all code points to random symbols
        for(u32 i=0;i<ncomb;i++)
            code_table[i]=rand()%nsymbols;    
        // Pre-fill low HW positions. They can still be altered, but this gives us a head start
        for(u32 i=0;i<numLowHammingPatterns;i++)
            code_table[idx2_1st_symbol_pattern[i]]=i%nsymbols; // Never exceed requested # of symbols
    }

    // Compute the initial score of the code table
    u32 score=computeCodeTableScore(code_table,numLowHammingPatterns);

    // Mark this score as the best one so far.
    setBest(score);

    // The "gap" is the distance to a solution, more convenient for monitoring
    printf("Initial gap: %d/%d\n",max_score-score,max_score);

    if(score>=max_score)
        return; // code_table holds a solution already

#ifdef PERFSTATS
    double perf_mut_count=0,perf_mut_ok=0;
    double perf_swap_count=0,perf_swap_ok=0;
#endif  

    u32 iterations_before_rebalance=rebalance_count;

    // Main iteration loop
    for(;;)
    {
        u32 old1=0,new1=0,p1=0,old2=0,new2=0,p2=0;
#ifndef EQUAL_WEIGHTS

        if(--iterations_before_rebalance==0) // Each rebalance_count iterations
        {
            setWeights(nsymbols1,weights); // Compute new weights for 1st write symbols
            score=computeCodeTableScore(code_table,numLowHammingPatterns); // Recompute entire score using the new weights
            iterations_before_rebalance=rebalance_count;
        }
#endif

#ifndef DONT_SAVE_INTERMEDIATES 
        // Conditionally save code_table - only when improved last save long enough in the past
        checkSave(nsymbols1, nsymbols2, nbit, score, code_table);
#endif
    
        // Apply a random mutation to the code table. A mutation type is picked at random from the following options:
        // - 'Point mutation': replace a single code point by different symbol
        // - 'Swap with neighbour': swap symbols of neighbouring code points (Hamming distance 1)
        // - (only if SWAP_ARBITRARY is defined) 'Swap random pair': swap symbols of arbitrary code points
        
#ifdef SWAP_ARBITRARY
        u32 muttype=rand()%3;
#else
        u32 muttype=rand()%2;
#endif
        if(muttype==0) // point mutation
        {
            p1=rand()%ncomb; // random code point
            old1=code_table[p1]; // save old symbol
            do
            {
                code_table[p1]=rand()%nsymbols; // new random symbol
            }while(code_table[p1]==old1); // Accept only true mutations
            
            new1=code_table[p1];
            updateCodeTableCoverage(code_table,p1,old1); // update coverage counters
#ifdef PERFSTATS            
            perf_mut_count++;
#endif
        }
        else // swap
        {
            p1=rand()%ncomb;// random code point
#ifdef SWAP_ARBITRARY           
            if(muttype==1)
                p2=rand()%ncomb; // 2nd random code point
            else // muttype==2
#endif
                p2= p1 ^ (1<<(rand()%nbit)); // p2 is a random NEIGHBOUR of p1
            
            old1=code_table[p1];
            old2=code_table[p2];
            
            u32 t=code_table[p1];
            code_table[p1]=code_table[p2];
            code_table[p2]=t;
            new1=code_table[p1];
            new2=code_table[p2];
            // update coverage counters
            updateCodeTableCoverage(code_table,p1,old1);
            updateCodeTableCoverage(code_table,p2,old2);
            
#ifdef PERFSTATS
            perf_swap_count++;
#endif
        }
        
        // Computing score of the mutant code table.
         
        u32 newscore=updateCodeTableScore(code_table,numLowHammingPatterns); 
        if(newscore>=score)
        {
            // Improved or identical score. Note that it is essential to accept identical scores:
            // it helps to maximize the amount of reachable states and fits in the general
            // philosophy of not getting stuck locally in the 2**nbit dimensional space.
#ifdef PERFSTATS
            if(muttype==0)
                perf_mut_ok++;
            else
                perf_swap_ok++;
#endif
            if(newscore>score)
            {
                score=newscore;
                printf("gap=%d/%d\n",max_score-score,max_score);
#ifdef PERFSTATS
                {
                    printf( "Mutations: %.0f ok / %.0f total\n", perf_mut_ok, perf_mut_count);
                    printf( "Swaps:     %.0f ok / %.0f total\n", perf_swap_ok, perf_swap_count);
                }
#endif
                if(score>=max_score)
                    return; // code_table holds a solution
            }
        }
        else 
        {
            // Worse situation than before the mutation. Revert the code table to the original
            code_table[p1]=old1;
            updateCodeTableCoverage(code_table,p1,new1);
            if(muttype>=1) // Revert from a swap
            {
                code_table[p2]=old2;
                updateCodeTableCoverage(code_table,p2,new2);
            }
            
        }
    
    }   

}



/**
 * For each code point, builds a list of potential 1st write code points from which this
 * point is reachable in the 2nd write.
 */

void initPrecursors()
{
    precursors = new Precursor_t[ncomb];
    u32 grandtotal=0;
    for(u32 p=0;p<ncomb;p++)
    {
        u32 count=0;
        // computing number of indices to allocate (! TODO: replace loop by formula!)
        for(u32 q=0;q<=p;q++)
        {
            if( (bitcount(q)<=first_write_threshold) && ((q&p)==q) )
                count++;
        }
        
        precursors[p].nprecursors = count;
        precursors[p].indices = new u32[count];
        grandtotal+=count;
        
        // count again after allocation
        count=0;
        for(u32 q=0;q<=p;q++) // TODO < or <= ?
        {
            if( (bitcount(q)<=first_write_threshold) && ((q&p)==q) )
                precursors[p].indices[count++]=pattern2_1st_symbol[q];
        }       
    }
    
    printf("Total of %u precursors\n",grandtotal);
    
}

// Create the coverage counters for each candidate 1st write symbol
void initCoverageCounters(u32 numLowHammingPatterns)
{
    Coveragecounters::SetDim(nsymbols2);
    coveragecounters=new Coveragecounters[numLowHammingPatterns];
}


int main(int argc, char *argv[])
{
    bool use_fixed_seed;
    fetchArguments(argc,argv,nsymbols1,nsymbols2,nbit,randomweight,rebalance_count,use_fixed_seed);
    ncomb=(1u<<nbit);
#ifndef EQUAL_WEIGHTS
    max_score=(nsymbols1+randomweight)*nsymbols2; // Bonus score distributed over 1st write symbols
#else
    max_score=nsymbols1*nsymbols2;
#endif
    nsymbols= (nsymbols1>nsymbols2) ? nsymbols1: nsymbols2;
    
    if(use_fixed_seed) // For benchmarking, makes results reproducible
    {
        srand(0x5EED);
    }
    else // Allow multiple sessions each having another chance to find a solution
    {
        srand(clock());
    }
    
    idx2_1st_symbol_pattern=new u32[ncomb]; // To be used to look up the patterns with low hamming count
    pattern2_1st_symbol=new u32[ncomb];
    memset(pattern2_1st_symbol,0,ncomb*sizeof(u32));
#ifndef EQUAL_WEIGHTS
    weights = new u32[nsymbols1];
    
    setWeights(nsymbols1,weights);
#endif
    
    for(first_write_threshold=0; first_write_threshold<nbit; first_write_threshold++)
    {   
        if(nsymbols2>(1u<<(nbit-first_write_threshold)))
        {
            printf("A <%u,%u>/%u two-write binary WOM code cannot exist.\n",nsymbols1,nsymbols2,nbit);
            return -1;
        }
        
        for(u32 i=0;i<ncomb;i++)
        {
            if(bitcount(i)==first_write_threshold)
            {
                pattern2_1st_symbol[i]=numLowHammingPatterns;
                idx2_1st_symbol_pattern[numLowHammingPatterns++]=i;
            }
        }
        
        if(numLowHammingPatterns>=nsymbols1)
        {
            break; // Sufficient amount of 1st write symbols reached. No need for additional bits in 1st write.
        }           
    }
    
    assert(numLowHammingPatterns>=nsymbols1);

    initPrecursors();
    initCoverageCounters(numLowHammingPatterns);

    code_table=new u32[ncomb];

    
    // Start the core routine: transform sequence into a valid solution
    findsolution(code_table, numLowHammingPatterns);
#ifndef DONT_SAVE_INTERMEDIATES
    saveState(nsymbols1, nsymbols2, nbit, code_table);
#endif
            
    // We think we found a solution, but let's check it anyway before we say hurray
    bool isok=solutionQualityCheck(nsymbols1, nsymbols2, nbit, code_table);
    if(isok)
    {
        storeSolution(nsymbols1, nsymbols2, nbit, code_table);
    }
    
    delete[] precursors;
    delete[] code_table;
#ifndef EQUAL_WEIGHTS
    delete[] weights;
#endif
    
    return 0;
}
