#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "support.h"

#define MAX_SYMBOLS 0xFFFF // Maximum number of primary or secondary write symbols
#define MAX_BITS 20        // Maximum code length
#define SAVE_INTERVAL 300u // Save every n seconds (if score is worth saving)
#define MAX_RANDOM_WEIGHT 100000
#define DEFAULT_REBALANCE_COUNT 65536

u32 bitcount(u32 pattern)
{
	pattern=(pattern&0x55555555)+((pattern>>1)&0x55555555);
	pattern=(pattern&0x33333333)+((pattern>>2)&0x33333333);
	pattern=(pattern&0x0F0F0F0F)+((pattern>>4)&0x0F0F0F0F);
	pattern+=pattern>>16;
	pattern+=pattern>>8;
	return pattern&0xFF;
}

static char *state_filename(u32 ns1, u32 ns2, u32 nbit)
{
	static char buf[100];
	sprintf(buf,"wom_finder_intermediate_state_%u_%u_%u.dat",ns1,ns2,nbit);
	return buf;
}

static char *solution_filename(u32 ns1, u32 ns2, u32 nbit)
{
	static char buf[100];
	sprintf(buf,"wom_finder_solution_%u_%u_%u.txt",ns1,ns2,nbit);
	return buf;
}


void saveState(u32 ns1, u32 ns2, u32 nbit, const u32 sequence[])
{
	const char *name = state_filename(ns1,ns2,nbit);
	FILE *f = fopen(name,"w");
	if(f!=NULL)
	{
		u32 ncomb=1<<nbit;
		fprintf(f,"# WOM_finder intermediate file in search for potential <%u,%u>/%u code\n#\n",ns1,ns2,nbit);
		fprintf(f,"# This file is saved periodically to back up work performed during long searches.\n");
		fprintf(f,"# When started with matching parameters, WOM_finder will load this file\n");
		fprintf(f,"# automatically at startup if it exists.\n#\n");
		for(u32 n=0;n<ncomb;n++)
		{
			fprintf(f,"%04x",sequence[n]);
			if((n%32)==31)
				fprintf(f,"\n");
		}
		fprintf(f,"\n");
		fclose(f);
		printf("Stored state file '%s'.\n",name);
	}
	else
	{
		perror("Could not save state file.");
	}
}

static u32 lastsavedscore=0;

void setBest(u32 bestscore)
{
	lastsavedscore=bestscore;
}


void checkSave(u32 ns1, u32 ns2, u32 nbit, u32 score, const u32 sequence[])
{
	static clock_t lastsavedtime=clock();
	if(score>lastsavedscore)
	{
		clock_t t=clock();
		 
		if( (t-lastsavedtime) >= (SAVE_INTERVAL * CLOCKS_PER_SEC) )
		{
			saveState(ns1,ns2,nbit,sequence);
			lastsavedtime=t;
			lastsavedscore=score;
		}
	}
}



bool attemptRestoreState(u32 ns1, u32 ns2, u32 nbit, u32 sequence[])
{
	const char *name = state_filename(ns1,ns2,nbit);
	FILE *f = fopen(name,"r");
	if(f==NULL)
		return false;
	printf("Restoring state from file '%s'.\n",name);
	u32 ncomb=1<<nbit;
	char buf[200];
	while(fgets(buf,sizeof(buf),f)!=NULL)
	{
		if(buf[0]=='#')
			continue;
		u32 nchars=strlen(buf);
		for(u32 offset=0;(ncomb>0) && (offset<(nchars-3));offset+=4)
		{
			u32 s=0;
			for(u32 k=0;k<4;k++)
			{
				char c=buf[offset+k];
				u32 v= ((c>='0')&&(c<='9')) ? (c-'0') : 10+(c-'a');
				if(v>15)
				{
					fprintf(stderr,"Error parsing state file.\n");
					return false;
				}
				
				s=(s<<4)+v;
			}
			
			*(sequence++)=s;
			ncomb--;
		}
	}
	
	
	if(ncomb!=0)
	{
		fprintf(stderr,"State file is incomplete or corrupted.\n");
	}
	
	return (ncomb==0);
}


bool solutionQualityCheck(u32 ns1, u32 ns2, u32 nbit, const u32 sequence[])
{
	const u32 ns=(ns1>ns2) ? ns1:ns2;
	const u32 ncomb=1<<nbit;
	
	if((ns>MAX_SYMBOLS)||(nbit>MAX_BITS))
	{
		fprintf(stderr, "ERROR - Input dimensions out of range.\n");
		return false;
	}
	
	bool coveredi[ns]={false};
	for(u32 i=0;i<ncomb;i++)
		if(!coveredi[sequence[i]])
		{
			bool coveredj[ns]={false};
			for(u32 j=0;j<ncomb;j++)
				if((~j & i)==0)
					coveredj[sequence[j]]=true;
			u32 j;
			
			for(j=0;j<ns2;j++)
				if(!coveredj[j])
					break;
			if(j==ns2)
				coveredi[sequence[i]]=true;
		}
	u32 i=0;
	for(i=0;i<ns1;i++)
		if(!coveredi[i])
		{
			fprintf(stderr, "ERROR - This shouldn't happen. NOK for symbol %u\n",i);
			return false;
		}
	
	printf("\nSuccess! Found a  <%d,%d>/%d  2-write binary WOM code.\n",ns1,ns2,nbit);
	return true;
}

void storeSolution(u32 ns1, u32 ns2, u32 nbit, const u32 sequence[])
{
	// Print to screen
	const u32 ncomb=1<<nbit;
	printf("\nAssignment of 2^%d symbols : {",nbit);
	for(u32 a=0;a<ncomb;a++)
	{
		if((a%32)==0)
			printf("\n");
		printf((a<(ncomb-1)) ? "%d," : "%d",sequence[a]);
		
	}
	printf("\n}\n\n");

	// Store to file also
	const char *name = solution_filename(ns1,ns2,nbit);
	FILE *f = fopen(name,"w");
	if(f!=NULL)
	{
		fprintf(f,"# WOM_finder solution for a <%u,%u>/%u code\n#\n",ns1,ns2,nbit);
		fprintf(f,"# Sum rate = %.6f\n", (log(ns1*ns2)/(nbit*log(2.0))));
		fprintf(f,"# The %u symbols listed below are mapped to code words that are the binary representation of the 0-based index in the table.\n",ncomb);
		
		for(u32 a=0;a<ncomb;a++)
		{
			if((a%32)==0)
				fprintf(f,"\n");
			fprintf(f,(a<(ncomb-1)) ? "%4d," : "%4d",sequence[a]);
			
		}
		
		fprintf(f,"\n# End of WOM_finder solution for a <%u,%u>/%u code\n",ns1,ns2,nbit);
		fclose(f);
		printf("Solution stored in '%s'\n",name);
		
	}
	else
	{
		perror("Could not open solution file.\n");
	}
}


static void usage()
{
	printf("\nWOM_finder [options] <ns1> <ns2> <nbits>\n");
	printf("\nTries to find a <ns1,ns2>/nbits  2-write binary WOM code.\n");
	printf("Options:\n");
	printf("    -w<randomweight>    :  Set random part of 1st symbol weights in score function.\n");
	printf("                           Default = 10 * ns1. '-w0' gives a deterministic score function.\n");
	printf("    -c<rebalance_count> :  Set number of iterations before weights are reassigned.\n");
	printf("                           Default = %u\n",DEFAULT_REBALANCE_COUNT);
	printf("    -d                  :  'Deterministic'. Uses fixed random seed for benchmarking purposes.\n\n");
	exit(1);
}

void fetchArguments(int argc, char *argv[], u32 &nsymbols1, u32 &nsymbols2, u32 &nbit, u32 &randomweight, u32 &rebalance_count, bool &use_fixed_seed)
{
	if(argc<4)
		usage();
	int idx=1;
	randomweight=0xFFFFFFFF;
	rebalance_count=DEFAULT_REBALANCE_COUNT;
	use_fixed_seed=false;
	
	while(idx<(argc-3))
	{
		if((strlen(argv[idx])>2)&&(argv[idx][0]=='-')&&(argv[idx][1]=='w'))
		{
			int ok=sscanf(&argv[idx++][2],"%u",&randomweight);
			if(ok!=1)
				usage();
		}
		else if((strlen(argv[idx])>2)&&(argv[idx][0]=='-')&&(argv[idx][1]=='c'))
		{
			int ok=sscanf(&argv[idx++][2],"%u",&rebalance_count);
			if((ok!=1)||(rebalance_count==0))
				usage();
		}
		else if (!strcmp(argv[idx++],"-d"))
		{
			use_fixed_seed=true;
		}
		else
		{
			usage();
		}
	}
	
	int r1=sscanf(argv[idx++],"%u",&nsymbols1);
	int r2=sscanf(argv[idx++],"%u",&nsymbols2);
	int r3=sscanf(argv[idx++],"%u",&nbit);
	
	if((r1!=1)||(r2!=1)||(r3!=1))
	{
		usage();
	}
	
	if((nsymbols1<2)||(nsymbols1>MAX_SYMBOLS))
	{
		printf("nsymbols1 out of range [2..%u]\n",MAX_SYMBOLS);
		exit(1);
	}
	if((nsymbols2<2)||(nsymbols2>MAX_SYMBOLS))
	{
		printf("nsymbols2 out of range [2..%u]\n",MAX_SYMBOLS);
		exit(1);
	}
	if((nbit<2)||(nbit>MAX_BITS))
	{
		printf("nbit out of range [2..%u]\n",MAX_BITS);
		exit(1);
	}
	if(randomweight==0xFFFFFFFF)
	{
		randomweight=10*nsymbols1; // Default random weight
	}
	
	if(randomweight>MAX_RANDOM_WEIGHT)
	{
		printf("randomweight out of range [0..%u]\n",MAX_RANDOM_WEIGHT);
		exit(1);
	}
	
}
